from typing import Sequence

from cleo.io.null_io import NullIO

# For fixing https://github.com/python-poetry/poetry/issues/5216
from packaging.tags import sys_tags  # noqa
from poetry.core.poetry import Poetry
from poetry.installation.operations.operation import Operation
from poetry.puzzle import Solver
from poetry.repositories import Pool
from poetry.repositories import Repository
from poetry.repositories.installed_repository import InstalledRepository
from poetry.utils.env import Env


def resolve_dependencies(poetry: "Poetry", env: Env, locked_repository: Repository) -> Sequence[Operation]:
    """
    This uses poetry's solver to resolve dependencies and filters out packages from the lock file which are not
    needed, such as packages that are not for our OS environment using markers (e.g. pywin32 is for Windows).
    """
    # Making a new repo containing the packages
    # newly resolved and the ones from the current lock file
    repo = Repository()
    for package in locked_repository.packages:
        if not repo.has_package(package):
            repo.add_package(package)

    pool = Pool(ignore_repository_names=True)
    pool.add_repository(repo)

    # Run through poetry's dependency resolver.  Only uses the default/main `dependencies` in pyproject.toml, not
    # dev dependencies or other groups.  If we want to support more groups in the wheel file, we can expand on the
    # CLI with more options.
    # See https://github.com/python-poetry/poetry/blob/master/src/poetry/installation/installer.py#L34 for poetry's
    # usage of this
    # TODO: Add support for adding more groups
    groups = ["default", "main"]

    solver = Solver(
        poetry.package.with_dependency_groups(groups=groups, only=True),
        pool,
        InstalledRepository.load(env),
        locked_repository,
        NullIO(),
    )

    # Everything is resolved at this point, so we no longer need
    # to load deferred dependencies (i.e. VCS, URL and path dependencies)
    solver.provider.load_deferred(False)

    with solver.use_environment(env):
        ops = solver.solve(use_latest=[]).calculate_operations(
            with_uninstalls=False,
            synchronize=False,
        )
    return ops



import subprocess
from typing import List

from cleo.helpers import argument, option
from cleo.io.inputs.option import Option

# For fixing https://github.com/python-poetry/poetry/issues/5216
from packaging.tags import sys_tags  # noqa
from poetry.console.commands.env_command import EnvCommand
from poetry.core.packages.package import Package

from poeblix.util import util


class ValidateDockerPlugin(EnvCommand):
    """
    Validates a docker container contains dependencies defined in the pyproject.toml and poetry.lock in the project
    this command is run.
    """

    name = "blixvalidatedocker"
    description = (
        "Validates a docker container contains dependencies as specified in pyproject.toml and poetry.lock.  "
        "By default, this validates in one direction, where dependencies specified in pyproject.toml/poetry.lock "
        "should be present in the Docker container, but not the other way around."
    )

    arguments = [argument("containerId", "Docker Container ID")]

    options: List[Option] = [
        option(
            "no-lock",
            None,
            "Disables validating lock file dependencies.",
        )
    ]

    loggers = ["poetry.core.masonry.builders.wheel"]

    def _validate_pyproject_toml(self, docker_deps: dict):
        cid = self.argument("containerId")
        required_packages = self.poetry.package.requires
        for package in required_packages:
            name = package.pretty_name
            if name in docker_deps:
                if package.accepts(Package(name, docker_deps[name])):
                    docker_deps.pop(name)
                else:
                    raise ValueError(
                        f"Inconsistency found!  pyproject.toml specifies {name}{package.constraint}, "
                        f"but docker container {cid} has {name}=={docker_deps[name]}"
                    )

    def _validate_poetry_lock(self, docker_deps: dict):
        if self.option("no-lock"):
            self.line("Skipping poetry.lock validation as --no-lock was specified")
            return

        cid = self.argument("containerId")
        locked_repo = self.poetry.locker.locked_repository()
        ops = util.resolve_dependencies(self.poetry, self.env, locked_repo)
        for op in ops:
            dependency_package = op.package
            name = dependency_package.pretty_name
            version = dependency_package.version
            if name in docker_deps:
                proc_version = docker_deps[name]
                if proc_version != str(version):
                    raise ValueError(
                        f"Inconsistency found!  poetry.lock specifies {name}=={version}, "
                        f"but docker container {cid} has {name}=={proc_version}"
                    )

    def handle(self) -> None:
        cid = self.argument("containerId")
        self.line(f"Fetching 'pip freeze' from docker image {cid} and validating against pyproject.toml/poetry.lock")

        docker_freeze = subprocess.check_output(f'docker exec {cid} python3 -m pip freeze | grep -v "@"', shell=True)
        docker_deps = {
            dsplit[0]: dsplit[1] for dsplit in (dep.split("==") for dep in docker_freeze.decode("utf-8").splitlines())
        }

        # Validate against pyproject.toml
        self._validate_pyproject_toml(docker_deps)
        # Validate against poetry.lock
        self._validate_poetry_lock(docker_deps)

        self.line(
            f"Validation success!  Docker image {cid} has consistent versions with dependencies specified "
            f"in this project"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['poeblix', 'poeblix.util']

package_data = \
{'': ['*']}

install_requires = \
['poetry>=1.2.0b1,<2.0']

entry_points = \
{'poetry.application.plugin': ['poeblix = poeblix.plugins:BlixPlugin']}

setup_kwargs = {
    'name': 'poeblix',
    'version': '0.3.1',
    'description': 'Poetry plugin that adds support for building wheel files using the poetry.lock file, and data_files just like in setup.py',
    'long_description': '# poeblix\nPoetry Plugin that adds various features deemed unfit for the official release, but makes sense to me\n\n# Overview\nThese contain custom poetry plugins that enable functionality not available in the official distribution of poetry.  These include:\n\n1. Using the Lock file to build a wheel file with pinned dependencies\n2. Support for data_files (like with setup.py) such as jupyter extensions or font files\n3. Validating a wheel file is consistent with dependencies specified in pyproject.toml/poetry.lock\n4. Validating a docker container\'s `pip freeze` contains dependencies as specified in pyproject.toml/poetry.lock\n\nThese are not supported in Poetry due to debate in the community: https://github.com/python-poetry/poetry/issues/890, https://github.com/python-poetry/poetry/issues/4013, https://github.com/python-poetry/poetry/issues/2778\n\n\n# How to Use\n\n### Prerequisite\n\nPoetry Plugins are only supported in 1.2.0+ which, at the moment (5/29/22), can only be installed when using the [new poetry installer](https://python-poetry.org/docs/master/#installation), and updating to the preview version via\n\n```commandline\npoetry self update --preview\n```\n\n## Installation\n\nYou can add the plugin via poetry\'s CLI:\n\n```commandline\npoetry plugin add poeblix\n```\n\nOr install directly from source/wheel, then add with the same above command using the path to the built dist\n\nTo update the plugin:\n\n```commandline\n# Update to latest\npoetry plugin add poeblix@latest\n\n# Update to specific version\npoetry plugin add poeblix==<version>\n```\n\n## Usage\n\n1. To build a wheel from your package (default uses poetry.lock to pin dependencies in the wheel):\n\n```commandline\npoetry blixbuild\n\n# To disable using lock file for building wheel\npoetry blixbuild --no-lock\n```\n\n2. Validate a wheel file has consistent dependencies and data_files as specified in pyproject.toml/poetry.lock\n\n```commandline\npoetry blixvalidatewheel <path-to-wheel>\n\n# Disable using lock file for validation\npoetry blixvalidatewheel --no-lock <path-to-wheel>\n```\n\n_Note: this validates consistency in both directions_\n\n3. Validate a docker container contains dependencies in a `pip freeze` as specified in pyproject.toml/poetry.lock\n\n```commandline\npoetry blixvalidatedocker <docker-container-ID>\n\n# Disable using lock file for validation\npoetry blixvalidatedocker --no-lock <docker-container-ID>\n```\n\n_Note: this only validates the docker container contains dependencies in the project, but not the other direction_\n\nHere\'s an example series of commands to start up a temporary docker container using its tag, validate it, then stop the temporary container\n\n```\n# This will output the newly running container id\ndocker run --entrypoint=bash -it -d <docker-image-tag>\n\n# Then validate the running docker container, and stop it when done\npoetry blixvalidatedocker <container-id>\ndocker stop <container-id>\n```\n\n4. Adding data_files to pyproject.toml to mimic data_files in setup.py:\n\n```yaml\n...\n\n[tool.blix.data]\ndata_files = [\n    { destination = "share/data/", from = [ "data_files/test.txt", "data_files/anotherfile" ] },\n    { destination = "share/data/threes", from = [ "data_files/athirdfile" ] }\n]\n\n...\n```\n\ndata_files should be under the `[tool.blix.data]` category and is a list of objects, each containing the `destination` data folder, and a `from` list of files to add to the destination data folder.\n\n_Note: the destination is a relative path that installs data to relative to the [installation prefix](https://docs.python.org/3/distutils/setupscript.html#installing-additional-files)_\n\nExample: https://github.com/spoorn/poeblix/blob/main/test/positive_cases/happy_case_example/pyproject.toml\n\n5. For more help on each command, use the --help argument\n\n```commandline\npoetry blixbuild --help\npoetry blixvalidatewheel --help\npoetry blixvalidatedocker --help\n```\n\n# Development\n\n```bash\n# Make a virtual environment on Python 3.9\n# If using virtualenvwrapper, run `mkvirtualenv -p python3.9 venv`\nvirtualenv -p python3.9 venv\n\n# Or activate existing virtualenv\n# If using virtualenvwrapper, run `workon venv`\nsource venv/bin/activate\n\n# installs the plugin in editable mode for easier testing via `poetry install`\n./devtool bootstrap\n\n# Lint checks\n./devtool lint\n\n# Tests\n./devtool test\n\n# Run all checks and tests\n./devtool all\n```\n\n**plugins.py** : contains our plugin that adds the `poetry blixbuild` command for building our wheel file\n\n**validatewheel.py**: adds a `poetry blixvalidatewheel` command that validates a wheel file contains the Required Dist as specified in pyproject.toml/poetry.lock\n\n**validatedocker.py** : adds a command that validates a docker file contains dependencies as specified in pyproject.toml and poetry.lock.  This does *NOT* validate that they are exactly matching, but rather that all dependencies in pyproject.toml/poetry.lock exist in the docker container on the correct versions.  The docker image may contain more extra dependencies\n',
    'author': 'spoorn',
    'author_email': 'spookump@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/spoorn/poeblix',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
